// Automatically generated file by cmake

#include "dart/common/Aspect.hpp"
#include "dart/common/AspectWithVersion.hpp"
#include "dart/common/ClassWithVirtualBase.hpp"
#include "dart/common/Cloneable.hpp"
#include "dart/common/Composite.hpp"
#include "dart/common/CompositeJoiner.hpp"
#include "dart/common/Console.hpp"
#include "dart/common/Deprecated.hpp"
#include "dart/common/EmbeddedAspect.hpp"
#include "dart/common/Empty.hpp"
#include "dart/common/Factory.hpp"
#include "dart/common/LocalResource.hpp"
#include "dart/common/LocalResourceRetriever.hpp"
#include "dart/common/LockableReference.hpp"
#include "dart/common/Memory.hpp"
#include "dart/common/NameManager.hpp"
#include "dart/common/Observer.hpp"
#include "dart/common/Optional.hpp"
#include "dart/common/Platform.hpp"
#include "dart/common/ProxyAspect.hpp"
#include "dart/common/RequiresAspect.hpp"
#include "dart/common/Resource.hpp"
#include "dart/common/ResourceRetriever.hpp"
#include "dart/common/SharedLibrary.hpp"
#include "dart/common/Signal.hpp"
#include "dart/common/Singleton.hpp"
#include "dart/common/SmartPointer.hpp"
#include "dart/common/SpecializedForAspect.hpp"
#include "dart/common/StlHelpers.hpp"
#include "dart/common/Subject.hpp"
#include "dart/common/Timer.hpp"
#include "dart/common/Uri.hpp"
#include "dart/common/VersionCounter.hpp"
#include "dart/common/Virtual.hpp"
#include "dart/common/sub_ptr.hpp"
